#ifndef HYPER_GENOM_MODEL_POSTER_HH_
#define HYPER_GENOM_MODEL_POSTER_HH_

#include <sstream>
#include <string>
#include <stdexcept>

#include <boost/optional.hpp>

#include <posterLib.h>

namespace hyper {
	namespace genom_model {
		class PosterException : public std::runtime_error {
			protected:
				std::string posterName;

			public:
				PosterException(const std::string& name) :
					std::runtime_error("PosterException"),
					posterName(name) {};
				virtual ~PosterException() throw() {};
				virtual const char* what () throw()  = 0;

				std::string errorMsg() const {
					char buf[1024];
					h2getErrMsg(errnoGet(), buf, sizeof(buf));
					return std::string(buf);
				};
		};

		class PosterFindException : public PosterException {
			public:
				PosterFindException(const std::string& name):
					PosterException(name) {};

				virtual const char* what() throw ()
				{
					std::ostringstream oss;
					oss << "Can't open " << posterName << " : " << errorMsg() << std::endl;
					return oss.str().c_str();
				}
		};

		class PosterCreateException : public PosterException {
			public:
				PosterCreateException(const std::string& name):
					PosterException(name) {};

				virtual const char* what() throw ()
				{
					std::ostringstream oss;
					oss << "Can't create " << posterName << " : " << errorMsg() << std::endl;
					return oss.str().c_str();
				}
		};

		template <typename T>
		class PosterReadException : public PosterException {
			private:
				int read;

			public:
				PosterReadException(const std::string& name, int err):
					PosterException(name), read(err) {};

				virtual const char* what() throw ()
				{
					std::ostringstream oss;
					oss << "Read " << read << " bytes ";
					oss << " : expected " << sizeof(T) << " bytes "; 
					oss << " : " << errorMsg() << std::endl;
					return oss.str().c_str();
				}
		};

		template <typename T>
		class PosterWriteException : public PosterException {
			private:
				int write;

			public:
				PosterWriteException(const std::string& name, int err):
					PosterException(name), write(err) {};

				virtual const char* what() throw ()
				{
					std::ostringstream oss;
					oss << "Wrote " << write << " bytes ";
					oss << " : expected " << sizeof(T) << " bytes "; 
					oss << " : " << errorMsg() << std::endl;
					return oss.str().c_str();
				}
		};

		template <typename T>
		class PosterReader
		{
			private:
				boost::optional<POSTER_ID> id;
				const std::string posterName;
				T value;

			private:
				void find(bool defered)
				{
					POSTER_ID c_id;
					STATUS s = posterFind(const_cast<char*>(posterName.c_str()), &c_id);
					if (s == ERROR && !defered) 
						throw PosterFindException(posterName);
					else if (s != ERROR)	id = c_id;
				}

			public:
				PosterReader(const std::string & name, bool defered = true) : 
					id(boost::none), posterName(name)
				{
					find(defered);
				}

				const T& operator() () 
				{
					if (!id) find(false);

					int err = posterRead(*id, 0, static_cast<void*>(&value), sizeof(T));
					if (err != sizeof(T))
						throw PosterReadException<T> (posterName, err);
					return value;
				}
		};

		template <typename T>
		class PosterWriter
		{
			private:
				POSTER_ID id;
				std::string posterName;

			public:
				PosterWriter(const std::string& name) : posterName(name) {
					STATUS s = posterCreate(const_cast<char*>(name.c_str()), sizeof(T), &id);
					if (s == ERROR) 
						throw PosterCreateException(name);
				}

				T& get() {
					T* addr = static_cast<T*> (posterAddr(id));
					return *addr;
				}

				void lock() { 
					posterTake(id, POSTER_WRITE);
				}

				void unlock() {
					posterGive(id);
				}

				void write(const T& value) {
					int err = posterWrite(id, 0, const_cast<void*>(static_cast<const void*>(&value)), 
												 sizeof(T));
					if (err != sizeof(T))
						throw PosterWriteException<T> (posterName, err);
				}

				~PosterWriter() { posterDelete(id); }
		};
	}
}

#endif /* HYPER_GENOM_MODEL_POSTER_HH_ */
